from random import random, sample

from .abs_Qstate import _Qstate
from .Basis import Basis, CanonBasis, stdbasis
from .Qbit import Qbit
from .Qbits import Qbits
from .Qerrors import IncomprehensibleStatusError, InitializationError
from .Qmath import ket, vector
from .qtils import STD_SYMBOLS, isScalar, states2list, str2states


#### Qstate.py
#
# This file contains the Qstate class, the "front" class for Qbit, Qbits, ecc...
# and the qbit function, a shortcut to create quantum states
#
####


# ↓↓↓↓↓↓↓↓↓↓↓↓ Qstate calss ↓↓↓↓↓↓↓↓↓↓↓↓ #


class Qstate(_Qstate):
    """
    Qstate is the quantum state class, it serves to create any kind of quantum states:
    
    + qudits (qubit, qutrit, etc), for example |0>
    + composed qudit states, for example |10010>

    Usage:
    
    + `state`: the vector that describes the quantum state
    + `basis` (optional): the basis of this Qstate, it serves to measure the Qstate and to represent it
    + `normalize` (optional): if True the state will be normalized
    + `transform` (optional): if True the state will be interpreted as a basis representation (so if state=ket(1,0) and basis={|+>,|->} the state will be interpreted as |+> and not |0>)
    """

    def __init__(self, state, basis=None, normalize=False, transform=False):
        
        if isinstance(state, vector):
            self.__class__ = Qbit
            self.__init__(state, basis, normalize, transform)
        
        else:
            self.__class__ = Qbits
            self.__init__(state, basis)


    @staticmethod
    def random(n = 2):
        "Generate a random quantum state of size `n`"
        if isinstance(n, Basis):
            b = n
            n = len(b)
        else:
            b = CanonBasis(n)
        rsign = lambda: sample((1,-1), 1)[0]
        return Qbit(ket([complex(rsign()*random(), rsign()*random()) for _ in range(n)]), b, normalize=True)


    @staticmethod
    def parse(s, basis = None, normalize = None):
        """
        Generate a quantum state parsing it from a string `s`
        
        The string must be formatted according to the classical representation of quantum states (bra-ket notation):  
        "+n1|s1> +n2|s2> ..." where `n` are complex numbers and `s` are the states of the basis `basis`
        
        Usage:
        + `s`: the string to be parsed
        + `basis` (optional): the Basis used in the `s` representation
        + `normalize` (optional): if True normalize the resulting state
        """
        try:
            state = str2states(s)
            if basis is None:
                if all((b in stdbasis.symbols for b in state)):
                    basis = stdbasis
                elif all((b in STD_SYMBOLS for b in state)):
                    basis = CanonBasis(STD_SYMBOLS.find(max(state.keys()))+1)
                elif len(state)>1:
                    basis = CanonBasis(len(state), list(state.keys()))
                else:
                    raise IncomprehensibleStatusError('Fail to parse')

            return Qbit(
                    ket(*states2list(state, basis.symbols)),
                    basis,
                    normalize=normalize,
                    transform=True
                )
            
        except Exception as e:
            raise InitializationError("Error to initialize Qbit from string", e)

# ↑↑↑↑↑↑↑↑↑↑↑↑ Qstate class ↑↑↑↑↑↑↑↑↑↑↑↑ #



def qbit(*something, basis=None, normalize=None, transform=False):
    "A shortcut to generate quantum states (see documentation for a complete description)"

    if len(something)>1:
        if isScalar(something[0]):
            return Qstate(ket(*something), basis, normalize=normalize, transform=transform)
        else:
            return Qstate(something, basis=basis)

    something = something[0] #single element
    if isinstance(something, vector):
        return Qstate(something, basis, normalize=normalize, transform=transform)

    elif isinstance(something, (list, tuple)):
        return Qstate(ket(*something), basis, normalize=normalize, transform=transform)

    elif isinstance(something, str):
        return Qstate.parse(something, basis, normalize=normalize)

    else:
        raise IncomprehensibleStatusError('Failed to understand which quantum state to generate')

qdit = qbit


from .abs_Qstate import _Qstate, unreal
from .Basis import Basis
from .Operator import MeasureOp, Op
from .Qerrors import IllegalOperationError, InitializationError
from .Qmath import ket, math, matrix, roundedVector
from .qtils import Vdigit, equal, formatProbs, mod_square, prod, val2str

#### Qbits.py
#
# This file contains the Qbits and Qent classes.
# These 2 classes serve to create composed quantum states,
# i.e. multiple qubit (or better: every kind of qudit) states, even entangled.
# 
# The Qent classes is a particular class that manage (hiddenly) the entangled states
#
####


# ↓↓↓↓↓↓↓↓↓↓↓↓ Common function ↓↓↓↓↓↓↓↓↓↓↓↓ #

def complete_Op(qbits, operator, pos):
    # it builds the operator to be applied, according to the given operator and the start position
    real_op = Op.neutral()
    if pos is None: pos=0
    
    if isinstance(operator, (list, tuple)): #set of operators
        if isinstance(pos, (list, tuple)):
            for i in range(len(qbits)):
                real_op @= operator[i] if i in pos else Op.Id(len(qbits[i]))
        else:
            for i in range(pos): real_op @= Op.Id(len(qbits[i]))
            for i in range(len(operator)): real_op @= operator[i]			

    else: #signle operators
        i=0
        while i<len(qbits):
            if i==pos:
                real_op @= operator
                l=1
                while l<len(operator):
                    l*=len(qbits[i])
                    i+=1
            else:
                real_op @= Op.Id(len(qbits[i]))
                i+=1
    
    return real_op


def isEnt(q):
    return q._ent is not None


def get_ent(qents):
    # get the Qbits_ent common to all Qbit in qents
    ent = qents[0]._ent
    if ent is None: return None
    for q in qents:
        if ent is not q._ent:
            ent=None
            break

    return ent


def gen_ent(qents):
    # if doesn't exist a common Qbits_ent, it generates it
    qent = get_ent(qents)
    if qent is None:
        qent = Qent(qents)

    return qent

# ↑↑↑↑↑↑↑↑↑↑↑↑ Common function ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Qbits class ↓↓↓↓↓↓↓↓↓↓↓↓ #

class Qbits(_Qstate):
    
    # Multiple qubit calss

    def __init__(self, qbits, basis=None):
        try:
            qs = []
            for i in range(len(qbits)):
                if isinstance(qbits[i], Qbits):
                    qs.extend(qbits[i]._qbits)
                else:
                    qs.append(qbits[i])
            
            if len(qs) != len(set(qs)):
                raise IllegalOperationError('Duplicate qubit not allowed')
            
            super().__init__(None, basis, qs, prod((len(q) for q in qs)))

        except Exception as e:
            raise InitializationError('Error to initialize the Qbits', e)


    def setBasis(self, basis = None):
        self._basis = basis

    
    def _getState(self):
        ent = get_ent(self._qbits)
        if ent is not None and len(ent) == len(self): #entangled state "equal" to this one
            self._qbits[0]._ent._prepare(self._qbits, lambda x,i: x[i]._pos)
            return self._qbits[0]._ent._state
        state = ket(1)
        for q in self._qbits:
            state = state @ q._getState()
        return state


    
    def __matmul__(self, q):
        if isinstance(q, int):
            raise IllegalOperationError("Entangled quantum states cannot be 'duplicated'")
        return Qbits(self._qbits + [q])



    def apply(self, operator, pos=None):
        if pos is not None or not isinstance(operator, (Op, MeasureOp)):
            operator = complete_Op(self._qbits, operator, pos)
        
        if operator._isSep():
            i=0
            for op in operator._pieces:
                if len(op) == len(self._qbits[i]):
                    self._qbits[i].apply(op)
                    i+=1
                else:
                    l=len(op) ; ent=[]
                    while(l>0):
                        ent.append(self._qbits[i])
                        l-=len(self._qbits[i]) ; i+=1
                    qent = gen_ent(ent)
                    qent._apply_qs(op, ent)
                    
        else:
            qent = gen_ent(self._qbits)
            qent._apply_qs(operator, self._qbits)
    


    def measure(self, basis = None):
        self._state = self._getState()
        i, state, basis = super().measure(basis)
        self._state = None

        self.apply(MeasureOp(state, basis, i))
        return basis.ew[i]


    def printAs(self, basis):
        isent = all((not isEnt(q) for q in self._qbits))
        state = basis.transform(self._getState())
        s = ''
        for i in range(len(self)):
            if(not equal(state[i], 0)):
                s += val2str(state[i]) + '|' + basis.symbols[i] + '> '
        
        if isent:
            s = s.replace('+', '±').replace('-', '±')
        
        return s[:-1]


    @unreal
    def printProbs(self, basis = None):
        b = basis if basis is not None else self._basis
        if b is None:
            T = matrix((1))
            for q in self._qbits:
                T @= q._basis
            state = roundedVector(T * self._getState())
            symb = Vdigit([q._basis.symbols for q in self._qbits])
        
        else:
            state = roundedVector(b.transform(self._getState()))
            symb = b.symbols
        
        print(formatProbs(state, symb))


    @unreal
    def __str__(self):
        if self._basis is None:
            T = matrix((1)) ; isent=False
            for q in self._qbits:
                T @= q.basis
                if isEnt(q): isent=True
            
            state = roundedVector(T*self._getState())
            s = "" ; st = Vdigit([q._basis.symbols for q in self._qbits])

            for e in state:
                val = next(st)
                if e != 0: s += val2str(e) + '|' + val + '> '
            
            if isent and get_ent(self._qbits) is None:
                s = s.replace('+', '±').replace('-', '±')
            
            return s[:-1]
        
        return self.printAs(self._basis)

# ↑↑↑↑↑↑↑↑↑↑↑↑ Qbits class ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Qent class ↓↓↓↓↓↓↓↓↓↓↓↓ #

class Qent(Qbits, _Qstate):
    
    # A particular class that manage the entangled states

    def __init__(self, qbits):
        try:
            qs_list = []
            state = ket(1)
            
            for qs in qbits:
                if isEnt(qs):
                    state @= qs._ent._state
                    for q in qs._ent._qbits:
                        q._entangle(self, len(qs_list))
                        qs_list.append(q)
                elif qs not in qs_list:
                    state @= qs._state
                    qs._entangle(self, len(qs_list))
                    qs_list.append(qs)
                    
            _Qstate.__init__(self, state, None, qs_list, len(state))
        
        except Exception as e:
            raise InitializationError('Error to initialize the Qbits entangled', e)


    def _getState(self):
        return self._state


    def apply(self, operator, pos=None):
        if pos is not None:
            index = self._prepare(pos)
            operator = complete_Op(self._qbits, operator, index)
        
        _Qstate.apply(self, operator)


    def apply2all(self, op):
        self.apply(op @ self.spaces)


    def _apply_qs(self, operator, qs):
        index = self._prepare(qs, get = lambda x,i: x[i]._pos)
        operator = complete_Op(self._qbits, operator, index)
        self.apply(operator)


    def measure(self, basis=None):
        i, _, basis = _Qstate.measure(self, basis)
        self._state = basis[i]
        return basis.ew[i]



    def _prepare(self, pos, get = lambda x,i: x[i]):
        if isinstance(pos, (list,tuple)):
            index = i = min((get(pos, i) for i in range(len(pos))))
            for p in range(len(pos)):
                j = get(pos, p)
                while i > j:
                    j = get(pos, j)
                self._swap(i, j)
                i+=1
            return index
        
        return pos


    def _swap(self, i, j):
        if i == j: return
        lens = [len(b) for b in self._qbits]
        b = min(len(self[i]), len(self[j]))
        state = Vdigit(lens)
        for x in range(b):
            for y in range(x+1, b):
                while state.hasNext:
                    state.ds[i] = x ; state.ds[j] = y
                    ind_i = state.extract()
                    state.ds[i] = y ; state.ds[j] = x
                    ind_j = state.extract()

                    #swap!
                    tmp= self._state[ind_i]
                    self._state[ind_i] = self._state[ind_j]
                    self._state[ind_j] = tmp

                    state._pp_ent(i,j,len(state.ds)-1)
        
        tmp = self._qbits[i]
        self._qbits[i] = self._qbits[j]
        self._qbits[j] = tmp
        #update
        self[i]._entangle(self, i)
        self[j]._entangle(self, j)
    

    def _calc_p_state(self, pos):
        #try to build the pos state using the probabilitiy to measure it
        state=[]
        for s in range(len(self._qbits[pos])):
            state.append(self._get_module(pos, s))
        return ket(*state)


    def _get_module(self, pos, state):
        # get the value of module of a certain state in a certain position
        d = 1
        for i in range(pos+1): d *= len(self._qbits[i])
        dim = len(self._state)//d
        step = dim*len(self._qbits[pos])
        val=0 ; sign=0
        for i in range(state*dim, len(self._state), step):
            for j in range(dim):
                sign += self._state[i+j]
                val += mod_square(self._state[i+j])

        return math.sqrt(val)

# ↑↑↑↑↑↑↑↑↑↑↑↑ Qent class ↑↑↑↑↑↑↑↑↑↑↑↑ #



from random import random

from .Basis import Basis
from .Operator import Op
from .Qerrors import IllegalOperationError, NotAllowError
from .Qmath import mod_square, vector


#### abs_Qstate.py
#
# This file contains the abstract class for Qstate
# and the functions to manage the auto-normalization and the cheating permissions
#
####


# ↓↓↓↓↓↓↓↓↓↓↓↓ Auto-normalization and cheating permissions functions ↓↓↓↓↓↓↓↓↓↓↓↓ #

_auto_norm = False
_cheat = True


def auto_norm(value = None):
    "If `True`, when a qubit is created it will be normalize if necessary"
    global _auto_norm
    _auto_norm = value if value in (False, True) else not _auto_norm


def cheat(value):
    "If `False`, disables the ability to see what is inside a qubit"
    if value in (False, True):
        global _cheat
        _cheat = value


def can_cheat():
    return _cheat


def need_norm():
    return _auto_norm


def unreal(func):
    # according to _cheat value, allow or not to launch the function that it decorates
    def wrapper(*args, **kwargs):
        if not _cheat:
            raise NotAllowError("Cheating is not allowed!")
        return func(*args, **kwargs)
    return wrapper

# ↑↑↑↑↑↑↑↑↑↑↑↑ Auto-normalization and cheating permissions functions ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ _Qstate class ↓↓↓↓↓↓↓↓↓↓↓↓ #


class _Qstate:
    # Abstract class for Qstate

    def __init__(self, state, basis, qbits, length):
        self._state = state
        self._basis = basis
        self._qbits = qbits
        self._length = length
        
    
    @property
    @unreal
    def state(self):
        return self._getState()
    

    @property
    def basis(self):
        return self._basis


    @property
    def spaces(self):
        "Return the number of qubits that compose this quantum state (i.e. the number of vectorial spaces)"
        return len(self._qbits)


    def __getitem__(self, i):
        return self._qbits[i]


    def _getState(self):
        return self._state


    def setBasis(self, basis):
        "Set the default basis for this quantum state"
        pass
    

    def apply(self, op, pos=None):
        "Apply an operator to this quantum state"
        self._state = op * self._state
        self._state.round_error((0,1))


    def apply2all(self, op):
        "Apply the operator to all the qubit that compose this quantum state"
        for q in self._qbits:
            q.apply(op)


    def measure(self, basis = None):
        "Measure this state in Basis `basis`, the result will be the eigenvalue associated"
        if basis is None:
            if self._basis is None: raise IllegalOperationError('Measurement basis required')
            basis = self._basis
        state = basis.transform(self._getState())
        r = random()
        i = -1 ; s = 1
        while s > r:
            i += 1
            s -= mod_square(state[i])
        return i, state, basis


    @unreal
    def prob(self, i, basis = None):
        "Returns the probability to measure the `i`-th state if the basis `basis` is used"
        if basis is None: basis = self._basis
        if basis is None: raise IllegalOperationError('Need a basis to calculate the probabilities')
        M = (basis[i] * ~basis[i])
        return ~self._getState() * (~M * M) * self._getState()


    @unreal
    def printProbs(self, basis = None):
        "Returns a string that describe the probabilities of measurement using basis `basis` of this quantum state"
        pass


    @unreal
    def printAs(self, basis):
        "Returns the representation of this quantum state from the 'point of view' of `basis`"
        pass


    def __len__(self):
        return self._length


    def __hash__(self):
        return id(self)
    

    def __str__(self):
        return self.printAs(self._basis)


    def __repr__(self):
        return self.__str__() if can_cheat() else '<Qstate @ '+hex(id(self))+', ('+self.__class__.__qualname__+')>'

# ↑↑↑↑↑↑↑↑↑↑↑↑ _Qstate class ↑↑↑↑↑↑↑↑↑↑↑↑ #



import math
from logiq import *


def deutch(f):
    Uf = Op.build({			#          f(0)=f(1)=0   f(0)=f(1)=1  f(0)=0,f(1)=1  f(0)=1,f(1)=0
            0: 0^f(0),		# |00>  -->    |00>          |01>          |00>          |01>
            1: 1^f(0),		# |01>  -->    |01>          |10>          |01>          |00>
            2: 2+0^f(1),	# |10>  -->    |10>          |11>          |11>          |10>
            3: 2+1^f(1)		# |11>  -->    |11>          |10>          |10>          |11>
        })
    B = Basis([[1,0],[0,-1]])
    q = qbit(1,0, basis=B) @ 2 #2 qubits having the same state: |00>

    Op.X | q[1]
    Op.H ^ q
    Uf | q
    Op.H | q[0]
    
    return q[0].measure(B)



def grover(db, f):
    vect = lambda x: ket(*(1 if i==x else 0 for i in range(len(db))))
    Uw = Op.build(
        {i: -vect(i) if f(i) else vect(i) for i in range(len(db))}
    )
    Us = Op(2*db.state * ~db.state - matrix.Id(len(db)))

    G = Uw*Us

    for i in range(int(round(math.sqrt(len(db))))):
        G | db

    db.measure(CanonBasis(len(db)))
    return db



from .Basis import Basis, CanonBasis, hadamard
from .Qerrors import DimensionError, InitializationError
from .Qmath import ket, kron, matrix, nkron, np, npmath, vector
from .qtils import Vdigit, find, isScalar, mod_square, states2list, str2states


#### Operator.py
#
# This file contains the Op class and the creation of most used operators
#
####


# ↓↓↓↓↓↓↓↓↓↓↓↓ Op classes ↓↓↓↓↓↓↓↓↓↓↓↓ #

class Op(matrix):
    """
    An Op (operator) server to modify the state of a quantum state

    + `operator`: a matrix that represent the operator
    """

    def __init__(self, operator, _pieces = None, no_cpy = False):
        try:
            if isinstance(operator, Op):
                super().__init__(operator.mtx, no_cpy=True)
                self._pieces = operator._pieces
            else:
                super().__init__(operator, no_cpy=no_cpy)
                if not self.isUnitary():
                    raise ValueError('Matrix not unitary')
            
            if _pieces is None:
                self._pieces = [self]
            else:
                self._pieces = _pieces
        
        except Exception as e:
            raise InitializationError("Error to initialize Op", e)
    

    def _isSep(self):
        return len(self._pieces) > 1


    def __or__(self, q):
        q.apply(self)

    
    def __xor__(self, q):
        q.apply2all(self)
    

    def __mul__(self, other):
        if isinstance(other, Op):
            return Op(self.mtx * other.mtx, no_cpy=True)

        elif isScalar(other) and mod_square(other)==1:
            return Op(self.mtx * other)

        return super().__mul__(other)
    

    def __rmul__(self, other):
        return super().__rmul__(other)


    def __matmul__(self, other):
        if isinstance(other, Op):
            return Op(kron(self.mtx, other.mtx), _pieces=self._pieces+other._pieces, no_cpy=True)

        elif isinstance(other, int):
            return Op(nkron(self, other), _pieces=self._pieces*other, no_cpy=True)

        return super().__matmul__(other)

    
    def __eq__(self, other):
        return npmath.equal(self, other)


    def __invert__(self):
         return Op(self.mtx.H)


    def __len__(self):
        return self.mtx.shape[0]


    @staticmethod
    def build(rules, basis = None):
        """
        Another way to create an operator is to describe it's behaviour with the autostate of a basis

        + `rules`: a dictionary {state : new_state} where state and new_state can be the vector, the number of autostate of basis or a string that represents the state (according to the basis symbols)
        + `basis` (optional): the basis which represents the rules (if it's the standard basis it's unnecessary)
        """

        op = matrix.filled(len(rules), 0)

        if basis is None:
            basis = CanonBasis(len(rules))
        elif isinstance(basis, (list,tuple)):
            #TODO: improve with __getitem__ in Vdigit && lazy init of CanonBasis
            symb = [str(v) for v in Vdigit(basis)]
            basis = CanonBasis(len(symb), symb)
        
        for k, v in rules.items():
            i = find(basis.symbols, k.strip('|> ')) if isinstance(k, str) else k
            
            if isinstance(v, str):
                state = basis.transform(ket(*states2list(str2states(v), basis.symbols)))
            elif isinstance(v, int):
                state = basis[v] if v>=0 else -basis[-v] #NB: beware of state 0: -0 is 0
            else:
                state = v

            op += state * ~basis[i]
                
        return Op(op.npm(), no_cpy=True)


    @staticmethod
    def neutral():
        "Returns an operator `N` such that `N @ x = x`"
        return Op([1], _pieces=[])
        #[[1]] @ x = x


    @staticmethod
    def Id(n):
        "Return an `n X n` identity operator"
        return Op(np.identity(n), no_cpy=True)


    @staticmethod
    def phaseGate(phase, deg = False):
        """
        Generates the 'phase gate' according to the giving `phase`  
        if `deg` is `True` the phase will be evaluated in degrees and not in radians"""
        if deg: phase = (2*np.pi*phase)/360.0
        return Op([[1, 0], [0, np.e**(1j * phase)]])


    @staticmethod
    def C(U):
        """
        The Controlled-U gate

        + `U`: the operator to apply to the second qubit if the first is "true"
        """
        if len(U) != 2: raise DimensionError('The operator U must be a 2x2 matrix')
        return Op([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, U[0,0], U[0,1]],
            [0, 0, U[1,0], U[1,1]]
            ])


    @staticmethod
    def random(n):
        "Generates a random `n X n` operator"
        return Op(matrix.rand_unitary(n), no_cpy=True)




class MeasureOp(matrix):

    #A special type of operator to permit the creation of a "measurement operator"
    #(because normally an operator can't be non-unitary)

    def __init__(self, state, basis, i):
        super().__init__(((basis[i] * ~basis[i])/state[i]).npm(), no_cpy=True)
        self._pieces = [self]


    def _isSep(self):
        return False


    def __or__(self, q):
        q.apply(self)


    def __xor__(self, q):
        q.apply2all(self)

    
    def __len__(self):
        return self.mtx.shape[0]


# ↑↑↑↑↑↑↑↑↑↑↑↑ Op classes ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Creation of most used Operators ↓↓↓↓↓↓↓↓↓↓↓↓ #

Op.I = Op( [[1,0],[0,1]] )
Op.X = Op( [[0,1],[1,0]] )
Op.Y = Op( [[0,-1j],[1j,0]] )
Op.Z = Op( [[1,0],[0,-1]] )
Op.H = Op(hadamard.npm(), no_cpy=True)
Op.cnot = Op( [[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]] )
Op.swap = Op( [[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]] )
Op.sqrtSwap = Op( [[1,0,0,0],[0,0.5*(1+1j),0.5*(1-1j),0],[0,0.5*(1-1j),0.5*(1+1j),0],[0,0,0,1]] )

# ↑↑↑↑↑↑↑↑↑↑↑↑ Creation of most used Operators ↑↑↑↑↑↑↑↑↑↑↑↑ #



from .src.abs_Qstate import auto_norm, cheat
from .src.Basis import Basis, CanonBasis, bell, hadamard, stdbasis
from .src.Operator import Op
from .src.Qmath import bra, ket, kron, matrix, nkron, roundedVector, vector
from .src.Qstate import Qstate, qbit, qdit
from .src.qtils import equal, equals, isScalar, set_n_digits, set_precision



from .Qerrors import InitializationError
from .Qmath import kron, math, matrix, np, roundedVector, vector
from .qtils import STD_SYMBOLS


#### Basis.py
#
# This file contains the Basis class and the creation of most used bases:
# - stdbasis: {|0>, |1>}
# - hadamard: {|+>, |->}
# - bell: {|Φ+>, |Ψ+>, |Ψ->, |Φ->}
#
####


# ↓↓↓↓↓↓↓↓↓↓↓↓ Basis classes ↓↓↓↓↓↓↓↓↓↓↓↓ #

class Basis(matrix):
    """
    A Basis serves to measuring and representing a quantum state

    + `basis`: a matrix that describe it or a list of eigenstate
    + `symbols` (optional): the symbols to represent the autosate of this new basis
    """

    def __init__(self, basis, symbols = None, no_cpy = False):
        try:
            if isinstance(basis, Basis):
                super().__init__(basis, no_cpy=True)
                self.symbols = basis.symbols
                self.ew = basis.ew
                return

            super().__init__(basis, no_cpy=no_cpy)

            if (not self.isOrthonormal()):
                raise ValueError("The vectors of the basis must be 'orthonormal' with each other")
            
            self.ew = tuple(np.linalg.eig(self.mtx)[0])

            if symbols != None:
                self.symbols = symbols
                if len(symbols) != len(self):
                    raise ValueError('Length of symbols and basis must be the same')
            elif len(self) <= len(STD_SYMBOLS):
                self.symbols = STD_SYMBOLS[:len(self)]
            else:
                raise ValueError('Too many states ('+str(len(self))+'), please specify a list of symbols')

        except Exception as e:
            raise InitializationError('Error to initialize Basis', e)
    

    def eigenstate(self, i):
        "returns the i-th eigenstate"       
        if isinstance(i, slice):
            return tuple(self.eigenstate(index) for index in range(*i.indices(len(self))))

        return super().__getitem__((slice(None), i))
    

    def __getitem__(self, i):
        return self.eigenstate(i)


    def __setitem__(self, i, value):
        raise NotImplementedError('Basis is an immutable object')
    

    def transform(self, vect):
        """
        Transform the vector `vect` into the one "saw from this basis"

        (i.e. apply the matrix that describe this basis to `vect`)
        """
        return roundedVector(self.mtx*vect if vect.isCol() else vect*self.T)


    def __or__(self, q):
        return q.measure(self)


    def __xor__(self, qs):
        c = []
        for q in qs:
            c.append(q.measure(self))
        return c


    def __matmul__(self, other):
        if isinstance(other, Basis):
            return Basis(kron(self, other), no_cpy=True)
        return super().__matmul__(other)

    
    def __invert__(self):
         return self
    

    def conj(self):
        raise NotImplementedError('Basis is an immutable object')
    

    def transpose(self):
        raise NotImplementedError('Basis is an immutable object')


    def __getattr__(self, name):
        if name in 'tT': return self
        elif name in 'hH': return self
        elif name == 'shape': return self.mtx.shape
        raise AttributeError("'Basis' object has no attribute '"+name+"'")

    
    def __len__(self):
        return self.shape[0]


    def __str__(self):
        s=''
        for i in range(len(self)):
            s += '|'+self.symbols[i]+'>: '+str(self[i])+'\n'
        return s
    
    def __repr__(self):
        return str(self)


    def measureOp(self, index, p):
        """
        Returns the operator that 'measures' the `index`-th state
        
        `p` is the probability to measure that state
        """
        return (1/p)*(self[index]*(self[index].t))


    @staticmethod
    def random(dim, symbols=None):
        "Return a random basis"
        return Basis(matrix.rand_orthonormal(dim), symbols=symbols, no_cpy=True)



class CanonBasis(Basis):
    """
    CanonBasis is a canonical basis (represented by an identical matrix)

    + `dim`: the dimension of this basis
    + `symbols` (optional): the symbols to represent the autosates of this new basis
    """

    def __init__(self, dim, symbols = None):
        if dim < 2:
            raise ValueError('Minimum length allow for a Basis is 2')
        if symbols is None:
            if dim > len(STD_SYMBOLS):
                symbols = dynSymb(dim)
            else:
                symbols = STD_SYMBOLS[:dim]

        super().__init__(np.identity(dim), symbols, no_cpy=True)
    
    
    def transform(self, vect: vector):
        return vect



class dynSymb:
    # simple class to provide illimitate symbols for big bases, without allocate useless memory

    def __init__(self, length):
        self.length = length


    def __getitem__(self, i):
        if 0<=i<self.length:
            return str(i)+' '
        raise IndexError() #TODO
    

    def __len__(self):
        return self.length


# ↑↑↑↑↑↑↑↑↑↑↑↑ Basis classes ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Creation of most used bases ↓↓↓↓↓↓↓↓↓↓↓↓ #

rd = 1/math.sqrt(2) #reciprocal diagonal ;)

stdbasis = CanonBasis(2)
hadamard = Basis(((rd,rd),(rd,-rd)), '+-')
bell = Basis(((rd,0,0,rd),(0,rd,rd,0),(0,rd,-rd,0),(rd,0,0,-rd)), ['Φ+','Ψ+','Ψ-','Φ-'])

# ↑↑↑↑↑↑↑↑↑↑↑↑ Creation of most used bases ↑↑↑↑↑↑↑↑↑↑↑↑ #



from .abs_Qstate import _Qstate, need_norm, unreal
from .Basis import Basis, CanonBasis
from .Operator import MeasureOp, Op
from .Qbits import Qbits
from .Qerrors import IllegalOperationError, InitializationError
from .Qmath import roundedVector, vector
from .qtils import equal, formatProbs, val2str


#### Qbit.py
#
# This file contains the Qbit and Qbit_ent classes.
# This 2 classes serve to create a pure quantum state even in an entangled state,
# i.e. a single qubit (or better: every kind of qudit)
#
####


# ↓↓↓↓↓↓↓↓↓↓↓↓ Qbit class ↓↓↓↓↓↓↓↓↓↓↓↓ #

class Qbit(_Qstate):
    
    # Single qubit class

    def __init__(self, state, basis = None, normalize = None, transform = False):
        try:
            super().__init__(vector(state), basis, [self], len(state))
            self._ent = None

            if not equal(self._state.norm(), 1):
                if (normalize is None and need_norm()) or normalize:
                    self._state.normalize()
                else:
                    raise IllegalOperationError('A quantum state must have norm 1, this one have norm '+str(self._state.norm()))
            
            if self._basis is None: self.setBasis(None)
            if transform: self._state = self._basis.transform(self._state)

        except Exception as e:
            raise InitializationError("Error to initialize Qbit", e)



    def setBasis(self, basis):
        if basis is None:
            self._basis = CanonBasis(len(self))
        else:
            self._basis = basis
            if len(self) != len(self._basis):
                raise ValueError('Lengths of state and basis must be equal')


    def _entangle(self, ent, pos):
        Qbit_ent.__init__(self, ent, pos, self._basis)
        self.__class__ = Qbit_ent


    def measure(self, basis = None):
        i, _, basis = super().measure(basis)
        self._state = vector(basis[i])

        return basis.ew[i]


    def __matmul__(self, q):
        if isinstance(q, int):
            return Qbits([Qbit(vector(self._state), self._basis) for _ in range(q)])
        return Qbits([self, q])


    @unreal
    def printAs(self, basis):
        s=''
        state = roundedVector(basis.transform(self._getState()))
        for i in range(len(self)):
            if(not equal(state[i], 0)):
                s += val2str(state[i]) + '|' + basis.symbols[i] + '> '

        return s[:-1]


    @unreal
    def printProbs(self, basis = None):
        b = basis if basis is not None else self._basis
        print(formatProbs(b.transform(self._getState()), b.symbols))

# ↑↑↑↑↑↑↑↑↑↑↑↑ Qbit class ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Qbit_ent class ↓↓↓↓↓↓↓↓↓↓↓↓ #

class Qbit_ent(Qbit, _Qstate):

    # Single qubit class for entangled states

    def __init__(self, ent, pos, basis):
        _Qstate.__init__(self, None, basis, [self], self._length)
        self._ent = ent
        self._pos = pos


    def _getState(self):
        return self._ent._calc_p_state(self._pos)


    def _entangle(self, ent, pos):
        self._ent = ent
        self._pos = pos


    def apply(self, op, pos = None):
        self._ent.apply(op, self._pos)


    def measure(self, basis = None):
        i, state, basis = _Qstate.measure(self, basis)
        self._ent.apply(MeasureOp(state, basis, i), self._pos)

        return basis.ew[i]



    def __matmul__(self, q):
        if isinstance(q, int):
            raise IllegalOperationError('This state is in entanglement')
        return Qbits([self, q])
    
    
    def printAs(self, basis):
        s='' ; rappr = '|>'
        state = roundedVector(basis.transform(self._getState()))
        for i in range(len(self)):
            if(not equal(state[i], 0)):
                s += val2str(state[i]) + rappr[0] + basis.symbols[i] + rappr[1] + ' '
        
        s = s.replace('+', '±').replace('-', '±')
        return s[:-1]

# ↑↑↑↑↑↑↑↑↑↑↑↑ Qbit_ent class ↑↑↑↑↑↑↑↑↑↑↑↑ #



import math
from logiq import *


rd=1/math.sqrt(2)

class sdc:
    #SuperDense Coding

    basis = Basis([
            [rd,0,0,rd],
            [0,rd*1j, rd*1j, 0],
            [0, rd*1j, -rd*1j, 0],
            [rd,0,0,-rd]
        ], '0231')


    def __init__(self, qbits):
        self.qs = qbits


    @staticmethod
    def init(n):
        n += 2 if n%2==0 else 1 #for padding!
        qs_a=[] ; qs_b=[]
        for _ in range(0,n,2):
            qa = qbit(1,0) ; qb = qbit(1,0)
            Op.H | qa
            q = qa @ qb
            Op.cnot | q
            qs_a.append(qa)
            qs_b.append(qb)
        return sdc(qs_a), sdc(qs_b)

    
    def prepare(self, bits): #prepare the qubits to send the bits
        # add padding
        bits.append(1)
        if (len(bits)-1)%2 == 0: bits.append(0)

        for i in range(0,len(bits), 2):
            if bits[i] == 0:
                if bits[i+1] == 0: #00
                    pass
                else: #01
                    Op.Z | self.qs[i//2]
            else:
                if bits[i+1] == 0: #10
                    Op.X | self.qs[i//2]
                else: #11
                    Op.Y*1j | self.qs[i//2]


    def receive(self, qs_a):
        bits=[]
        for i in range(len(qs_a)):
            m = (qs_a[i] @ self.qs[i]).measure(sdc.basis)
            if m == sdc.basis.ew[0]: bits+=[0,0]
            elif m == sdc.basis.ew[1]: bits+=[1,0]
            elif m == sdc.basis.ew[2]: bits+=[1,1]
            else: bits+=[0,1]
        
        # remove padding
        while bits.pop() == 0: pass

        return bits


    def __getitem__(self,i):
        return self.qs[i]

    def __len__(self):
        return len(self.qs)

    def __repr__(self):
        return str(self.qs)



from random import random, sample

from .Qerrors import DimensionError, GenericLogiqError, InitializationError
from .qtils import equal, isScalar, math, mod_square, np


#### Qmath.py
#
# This file contains 2 classes that wrap the numpy.matrix class: the vector and the matrix class,
# it contains also other mathematical functions used in the classes.
# (In the future, these 2 classes could be refactored and became the subclasses of a superclass "linobj")
#
####


# ↓↓↓↓↓↓↓↓↓↓↓↓ Add the npm() method to the np.matrix class ↓↓↓↓↓↓↓↓↓↓↓↓ #
def _npm(self):
    return self

setattr(np.matrix, 'npm', _npm)
# ↑↑↑↑↑↑↑↑↑↑↑↑ Add the npm() method to the np.matrix class ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Kronecker product functions ↓↓↓↓↓↓↓↓↓↓↓↓ #

def kron(m1, m2):
    "Return the Kronecker product between `m1` and `m2`  \n(`m1` and `m2` must be `vector` or `matrix`, otherwise they can be a `np.matrix`)"
    try: npm1 = m1.npm() ; npm2 = m2.npm()
    except AttributeError: raise TypeError("The types of the inputs must be vector or matrix")
    return select_type(np.kron(npm1, npm2))


def nkron(m, n):
    "Apply the Kronecker product `n` times: `(m @ m @ ... @ m)` for `n` times"
    if n<1: return matrix((1)) #[[1]] @ x = x
    out = m
    for _ in range(n-1):
        out = kron(out, m)
    return out

# ↑↑↑↑↑↑↑↑↑↑↑↑ Kronecker product functions ↑↑↑↑↑↑↑↑↑↑↑↑ #


# ↓↓↓↓↓↓↓↓↓↓↓↓ Other useful functions ↓↓↓↓↓↓↓↓↓↓↓↓ #

def select_type(item):
    if len(item.shape) == 0:
        return item
    elif item.shape == (1,1):
        return item[0,0]				#scalar
    elif min(item.shape) == 1:
        return vector(item, no_cpy=True)#vector
    else:
        return matrix(item, no_cpy=True)#matrix



class npmath:
    # Methods to use, safely, standard operators between scalars, vector, matrix and numpy.matrix

    @staticmethod
    def safe_npm(x):
        try: return x if isScalar(x) else x.npm()
        except: raise TypeError("'"+str(type(x))+"' is an invalid type")
        
    @staticmethod
    def _fun(x, y, fun):
        return select_type( fun(npmath.safe_npm(x), npmath.safe_npm(y)) )

    @staticmethod
    def add(x, y):
        return npmath._fun(x, y, lambda x,y: x+y)

    @staticmethod
    def sub(x, y):
        return npmath._fun(x, y, lambda x,y: x-y)

    @staticmethod
    def mul(x, y):
        return npmath._fun(x, y, lambda x,y: x*y)

    @staticmethod
    def div(x, y):
        return npmath._fun(x, y, lambda x,y: x/y)

    @staticmethod
    def equal(x, y):
        if isScalar(x) and isScalar(y): return x == y
        try: x = x.npm() ; y = y.npm()
        except: return False
        return npmath.np_equal(x, y) #NB: missing case "isScalar(x) or isScalar(y)"

    @staticmethod
    def np_equal(x, y):
        if x.shape != y.shape: return False
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                if x[i,j] != y[i,j]: return False
        return True

# ↑↑↑↑↑↑↑↑↑↑↑↑ Other useful functions ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ vector class ↓↓↓↓↓↓↓↓↓↓↓↓ #

class vector:
    """
    + `v`: describes the value of this vector, it can be every kind of object allowed from `numpy.matrix` constructor, or another vector (in that case it will be copied)
    + `normalize` (optional): if True this vector will normalize (i.e. his norm became 1)
    + `values2round` (optional): a list where the value inside will be rounded for example if `values2round=[1]` and the vector is `[1.00001, 12.00001]` it becomes `[1, 12.00001]` (this according to the global precision)
    """

    def __init__(self, v, normalize = False, values2round = None, no_cpy = False):
        try:
            if no_cpy and isinstance(v, np.matrix):
                self.vect = v
            elif isinstance(v, (vector, matrix)):
                self.vect = np.matrix(v.npm())
            else:
                self.vect = np.matrix(v, complex)
            
            if min(self.vect.shape) != 1:
                raise DimensionError('A vector must be have only one dimension')
        
        except Exception as e:
            raise InitializationError('Error to initialize the vector', e)

        if normalize: self.normalize()
        if values2round is not None: self.round_error(values2round)


    def npm(self):
        """
        Return the reference of the `numpy.matrix` associated
        Watch out: it return the reference of the vector (for performance reason), so if you want to modify it first copy it.
        """
        return self.vect


    def isRow(self):
        "`True` if the vector is a 'row vector'"
        return self.vect.shape[0] == 1
    

    def isCol(self):
        "`True` if the vector is a 'column vector'"
        return self.vect.shape[1] == 1
    

    def round_error(self, values):
        """Round the error (using `equals()`) of values in `values`.\n
        For example if `values = (0,1)` and this vector is `|5, 0.9999998>` it may became `|5, 1>`"""
        for i in range(len(self)):
            for v in values:
                if equal(self[i], v):
                    self[i] = v
                elif equal(self[i].real, 0) or equal(self[i].imag, 0):
                    self[i] = complex( 0 if equal(self[i].real, 0) else self[i].real, 0 if equal(self[i].imag, 0) else self[i].imag )


    def norm(self):
        "Return the norm of this vector, defined as `v•vt` where `vt` is the transpose of `v`"
        return ~self * self if self.isCol() else self * ~self
    

    def norm2(self):
        "Return the norm 2 of this vector"
        s = 0
        for e in self:
            s += mod_square(e)
        return math.sqrt(s)
    

    def normalize(self):
        "Normalize this vector (i.e. his norm became 1)"
        d = self.norm2()
        if d==0: raise DimensionError("The null vector isn't normalizable")
        for i in range(len(self)):
            self[i] /= d
        return self


    def __pos__(self):
        pass


    def __neg__(self):
        return vector(self.vect.__neg__(), no_cpy=True)


    def __len__(self):
        return max(self.vect.shape)
    

    def __getitem__(self, i):
        try:
            item = self.vect[0, i] if self.isRow() else self.vect[i, 0]
            return vector(item) if isinstance(i, slice) else item
        except IndexError:
            raise IndexError("Index out of bound (i="+str(i)+" and vector's length = "+str(len(self))+")")
    

    def __setitem__(self, i, value):
        try:
            if self.isRow(): self.vect[0, i] = value
            else: self.vect[i, 0] = value
        except IndexError:
            raise IndexError("Index out of bound (i="+str(i)+" and vector's length = "+str(len(self))+")")
    

    def _vector_op(self, v, fun):
        #BUG: numpy.complex128 + v = np.array
        if isinstance(v, vector):
            if v.vect.shape != self.vect.shape: raise DimensionError("The vector's lengths must be equal")
            return vector(fun(self.vect, v.vect), no_cpy=True)

        elif isScalar(v):
            return vector(fun(self.vect, v), no_cpy=True)

        else: raise TypeError('Other member must be a vector or a scalar')
        

    def __add__(self, v):
        try:
            return self._vector_op(v, lambda x,y: x+y)
        except Exception as e:
            raise GenericLogiqError('Unable to sum', e)
    

    def __radd__(self, v):
        try:
            return self._vector_op(v, lambda x,y: y+x)
        except Exception as e:
            raise GenericLogiqError('Unable to sum', e)
    

    def __sub__(self, v):
        try:
            return self._vector_op(v, lambda x,y: x-y)
        except Exception as e:
            raise GenericLogiqError('Unable to subtract', e)
    

    def __rsub__(self, v):
        try:
            return self._vector_op(v, lambda x,y: y-x)
        except Exception as e:
            raise GenericLogiqError('Unable to subtract', e)

    
    def __mul__(self, v):
        npv = npmath.safe_npm(v)
        return select_type(self.vect * npv)
    

    def __rmul__(self, v):
        npv = npmath.safe_npm(v)
        return select_type(npv * self.vect)
    

    def __truediv__(self, v):
        npv = npmath.safe_npm(v)
        return select_type(self.vect / npv)


    def __rtruediv__(self, v):
        npv = npmath.safe_npm(v)
        return select_type(npv / self.vect)
    

    def __matmul__(self, v):
        if isinstance(v, int): return nkron(self, v)
        return kron(self, v)
    

    def __eq__(self, v):
        if isinstance(v, (vector, matrix, np.matrix)):
            return npmath.np_equal(self.vect, v.npm())
        else:
            return False


    def __invert__(self): #conjugate transpose
         return vector(self.vect.H)
    

    def transpose(self):
        "Transform this vector into its transpose"
        self.vect = self.vect.T
    
    
    def conj(self):
        "Transform this vector into its conjugate transpose"
        self.vect = self.vect.H


    def __getattr__(self, name):
        if name in 'tT': return vector(self.vect.T)
        elif name in 'hH': return ~self
        elif name == 'shape': return self.vect.shape
        raise AttributeError("'vector' object has no attribute '"+name+"'")
    

    def __hash__(self):
        return hash(str(self.vect)) #fast hash


    def __str__(self):
        s=''
        for i in range(len(self)): s += str(self[i])+'; '
        return ('|' if self.isCol() else '<') + s[:-2] + ('>' if self.isCol() else '|')
    

    def __repr__(self):
        return str(self)


    @staticmethod
    def random(dim):
        "Return a random complex vector long `dim`"
        rsign = lambda: sample((1,-1), 1)[0]
        return vector([complex(rsign()*random(), rsign()*random()) for _ in range(dim)])

# ↑↑↑↑↑↑↑↑↑↑↑↑ vector class ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Useful vector constructors ↓↓↓↓↓↓↓↓↓↓↓↓ #

def roundedVector(v, values2round=(0,1,-1)):
    "Generate a vector with the values in `values2round` rounded"
    return vector(v, values2round=values2round)



class ket(vector):
    """A shortcut to create a column vector
    
    Usage: `ket(x0, x1, ..., xn)`"""

    def __init__(self, *values):
        super().__init__(values)
        if len(values) == 1: values = values[0]
        if self.isRow():
            if isinstance(values, vector): self.conj() #transformation from ket to bra
            else: self.transpose() #no transformation, only creating a vector column



class bra(vector):
    """A shortcut to create a row vector
    
    Usage: `bra(x0, x1, ..., xn)`"""

    def __init__(self, *values):
        if len(values) == 1: values = values[0]
        super().__init__(values)
        if self.isCol():
            if isinstance(values, vector): self.conj() #transformation from ket to bra
            else: self.transpose() #no transformation, only creating a vector column

# ↑↑↑↑↑↑↑↑↑↑↑↑ Useful vector constructors ↑↑↑↑↑↑↑↑↑↑↑↑ #


# ↓↓↓↓↓↓↓↓↓↓↓↓ matrix class ↓↓↓↓↓↓↓↓↓↓↓↓ #

class matrix:
    """
    + `M`: describes the value of this vector, it can be every kind of object allowed to `numpy.matrix` constructor, or another matrix (in that case it will be copied)
    + `values2round` (optional): a list where the value inside will be rounded, for example if `values2round=[1]` and the matrix is `[[1.00001, 12.00001],[0.3,1.12]]` it becomes `[[1, 12.00001],[0.3,1.12]]` (this according to the global precision)
    """
    
    def __init__(self, M, values2round = None, no_cpy = False):
        try:
            if no_cpy and isinstance(M, np.matrix):
                self.mtx = M
            elif isinstance(M, (matrix, vector)):
                self.mtx = np.matrix(M.npm())
            else:
                self.mtx = np.matrix(M, complex)
        except Exception as e:
            raise InitializationError('Error to initialize the matrix', e)

        if values2round is not None: self.round_error(values2round)


    def round_error(self, values):
        """Round the error (using `equals()`) of values in `values`.  
        For example if `values = (0,1)` and this matrix is `[[5, 0.9999998],[-0.000001, 0.1]]` it may became `[[5, 1],[0, 0.1]]`"""
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                for v in values:
                    if equal(self[i, j], v): self[i, j] = v


    def isUnitary(self):
        "`True` if this matrix is unitary"
        return self.mtx.shape[0] % self.mtx.shape[1] == 0 and np.allclose(np.eye(self.mtx.shape[0]), self.mtx.H * self.mtx)


    def isOrthonormal(self):
        "`True` if this matrix is orthonormal"
        _, r = np.linalg.qr(self.mtx)
        for i in range(len(self)):
            for j in range(i, len(self)):
                if i==j: continue
                elif not equal(r[i,j], 0):
                    return False
        return True


    def nomr(self):
        "return the norm of this matrix"
        return np.linalg.norm(self.mtx)
    

    def det(self):
        "Return the determinant of this matrix"
        return np.linalg.det(self.mtx)


    def npm(self):
        """
        Return the reference of the `numpy.matrix` associated.  
        Watch out: it return the reference of the matrix (for performance reason), so if you want to modify it first copy it.
        """
        return self.mtx


    def __pos__(self):
        pass


    def __neg__(self):
        return matrix(self.mtx.__neg__(), no_cpy=True)


    def __getitem__(self, i):
        try:
            return select_type(self.mtx[i])
        except IndexError:
            raise IndexError("Index out of bound (request="+str(i)+", shape="+str(self.shape)+")")


    def __setitem__(self, i, value):
        self.mtx[i] = value
    

    def _matrix_op(self, M, fun):
        #BUG: numpy.complex128 + M = np.array
        if isinstance(M, matrix):
            if M.shape != self.shape: raise DimensionError("The matrices' shapes must be equal")
            return matrix(fun(self.mtx, M.mtx), no_cpy=True)

        elif isScalar(M):
            return matrix(fun(self.mtx, M), no_cpy=True)

        else: raise TypeError('Other member must be a matrix or a scalar')
        

    def __add__(self, M):
        try:
            return self._matrix_op(M, lambda x,y: x+y)
        except Exception as e:
            raise GenericLogiqError('Unable to add', e)


    def __radd__(self, M):
        try:
            return self._matrix_op(M, lambda x,y: y+x)
        except Exception as e:
            raise GenericLogiqError('Unable to add', e)


    def __sub__(self, M):
        try:
            return self._matrix_op(M, lambda x,y: x-y)
        except Exception as e:
            raise GenericLogiqError('Unable to subtract', e)


    def __rsub__(self, M):
        try:
            return self._matrix_op(M, lambda x,y: y-x)
        except Exception as e:
            raise GenericLogiqError('Unable to subtract', e)
    

    def __mul__(self, M):
        npm = npmath.safe_npm(M)
        return select_type(self.mtx * npm)
    

    def __rmul__(self, M):
        npm = npmath.safe_npm(M)
        return select_type(npm * self.mtx)
    

    def __truediv__(self, M):
        npm = npmath.safe_npm(M)
        return select_type(self.mtx / npm)


    def __rtruediv__(self, M):
        npm = npmath.safe_npm(M)
        return select_type(npm / self.mtx)


    def __matmul__(self, M):
        if isinstance(M, int): return nkron(self, M)
        return kron(self, M)


    def __eq__(self, M):
        if isinstance(M, (vector, matrix, np.matrix)):
            return npmath.np_equal(self.npm(), M.npm())
        else:
            return False


    def __invert__(self):
         return matrix(self.mtx.H)
    

    def transpose(self):
        "Transform this matrix into its transpose"
        self.mtx = self.mtx.T


    def conj(self):
        "Transform this matrix into its conjugate transpose"
        self.mtx = self.mtx.H


    def __getattr__(self, name):
        if name in 'tT': return matrix(self.mtx.T)
        elif name in 'hH': return ~self
        elif name == 'shape': return self.mtx.shape
        raise AttributeError("'matrix' object has no attribute '"+name+"'")


    def __hash__(self):
        return hash(str(self.mtx)) #fast hash


    def __str__(self):
        return str(self.mtx)
    

    def __repr__(self):
        return str(self)


    @staticmethod
    def Id(n):
        "Return an `n X n` identity matrix"
        return matrix(np.identity(n), no_cpy=True)


    @staticmethod
    def filled(shape, val=0):
        """Creates a matrix `shape[0] X shape[1]` filled with the value in `val`  
        NB: if shape is a natural number, the matrix will be a square matrix `shape X shape`"""
        if isinstance(shape, int): shape = (shape, shape)
        if val == 0:
            return matrix(np.zeros(shape, complex), no_cpy=True)
        elif val == 1:
            return matrix(np.ones(shape, complex), no_cpy=True)
        else:
            M = matrix(np.empty(shape, complex), no_cpy=True)
            for i in range(M.shape[0]):
                for j in range(M.shape[0]):
                    M[i,j] = val
            return M
    

    @staticmethod
    def random(shape):
        "Return a random complex matrix according to the given shape"
        if isinstance(shape, int): shape = (shape, shape)
        rsign = lambda: sample((1,-1), 1)[0]
        return matrix([[complex(rsign()*random(), rsign()*random()) for _ in range(shape[0])] for _ in range(shape[1])])


    @staticmethod
    def rand_unitary(dim):
        random_state = np.random
        z = 1/math.sqrt(2)*(random_state.normal(size=(dim, dim)) + 1j*random_state.normal(size=(dim, dim)))
        q, r = np.linalg.qr(z)
        d = r.diagonal()
        q *= d/abs(d)
        return matrix(q, no_cpy=True)


    @staticmethod
    def rand_orthonormal(dim):
        random_state = np.random
        H = np.eye(dim)
        D = np.ones((dim,))
        for n in range(1, dim):
            x = random_state.normal(size=(dim-n+1,))
            D[n-1] = np.sign(x[0])
            x[0] -= D[n-1]*np.sqrt((x*x).sum())
            # Householder transformation
            Hx = (np.eye(dim-n+1) - 2.*np.outer(x, x)/(x*x).sum())
            mat = np.eye(dim)
            mat[n-1:, n-1:] = Hx
            H = np.dot(H, mat)
            # Fix the last sign such that the determinant is 1
        D[-1] = (-1)**(1-(dim % 2))*D.prod()
        # Equivalent to np.dot(np.diag(D), H) but faster, apparently
        H = (D*H.T).T
        return matrix(H, no_cpy=True)


# ↑↑↑↑↑↑↑↑↑↑↑↑ matrix class ↑↑↑↑↑↑↑↑↑↑↑↑ #



from setuptools import find_packages, setup

with open('LICENSE', encoding='utf-8') as f:
    LICENSE = f.read()

setup(
    name='logiq',
    version='0.1.1',
    description='Logiq, a simple to use quantum simulator library',
    long_description='Github repo: https://github.com/Bnz-0/logiq',
    author='Bnz',
    author_email='matteo.benzi97@gmail.com',
    url='https://github.com/Bnz-0/logiq',
    license=LICENSE,
    python_requires='>=3.6.0',
    packages=find_packages(exclude=('tests', 'docs', 'TODO')),
    install_requires=['numpy']
)



#### Qerrors.py
#
# This file contains all type of custom exception that can be raised in logiq
#
####


class_name = lambda obj: '<'+obj.__class__.__name__+'>'

class LogiqError(Exception):

    def __init__(self, mex, prev=None):
        self.mex = mex
        self.prev = prev
    

    def __str__(self):
        return self.mex + ('' if self.prev is None else ', raised by ' + class_name(self.prev) + ': ' + str(self.prev))


class GenericLogiqError(LogiqError):
    "Generically raised when something goes wrong"
    pass

class InitializationError(LogiqError):
    "Raised when an object failed to initialize itself"
    pass

class DimensionError(LogiqError):
    "Raised when a linear object's dimension isn't compatible"
    pass

class NotAllowError(LogiqError):
    "Raised when the operation is not allowed"
    pass

class IllegalOperationError(LogiqError):
    "Raised when the simulator system can't do something"
    pass

class IncomprehensibleStatusError(LogiqError):
    "Raised when a status of a quantum state, described in some way, contain a \"sintax\" error"
    pass



import math
import re
from collections import defaultdict
from numbers import Complex

import numpy as np


#### qtils.py
#
# This file contains some utility functions, in particular:
# - to manage the precision and provide functions that kepp in mind it
# - to formatting the qtrings of qubit representation
# - the Vdigit class
# - other
#
####


# standard symbols used for representations
STD_SYMBOLS = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'


def prod(l):
    p=1
    for e in l: p *= e
    return p


def find(l, e):
    for i in range(len(l)):
        if e == l[i]: return i
    return None


# ↓↓↓↓↓↓↓↓↓↓↓↓ Precision ↓↓↓↓↓↓↓↓↓↓↓↓ #

current_prec = DEFAULT_PRECISION = np.finfo(np.complex128).precision -1


def set_precision(p = None):
    "Set the precision used by the function `equal()`"
    global current_prec
    if p is None: current_prec = DEFAULT_PRECISION
    elif p > 0:
        current_prec = p
    else:
        raise ValueError('The value of precision must be an int greater than 0')


def what_precision():
    "Returns the current precision"
    return current_prec

# ↑↑↑↑↑↑↑↑↑↑↑↑ Precision ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Necessary mathematic functions ↓↓↓↓↓↓↓↓↓↓↓↓ #

def mod_square(c):
    "Returns the module square of a complex number, which is defined as c•c*, where c* is the complex conjugate of `c`"
    return (c * c.conjugate()).real


def equal(n1, n2):
    "Check if 2 complex numbers are equal, keeping in mind possible error of approximation."
    if isinstance(n1, np.inexact):
        return np.around(n1, current_prec) == np.around(n2, current_prec)
    else:
        return n1 == n2


def equals(v1, v2):
    "Check if 2 linear objects are equal, keeping in mind possible error of approximation"
    try: npm1 = v1.npm() ; npm2 = v2.npm()
    except AttributeError: return v1 == v2
    if npm1.shape != npm2.shape: return False
    for i in range(npm1.shape[0]):
        for j in range(npm1.shape[1]):
            if not equal(npm1[i,j], npm2[i,j]): return False
    return True


def isScalar(var):
    "True if `var` is a scalar (i.e. a value in the complexes)"
    return isinstance(var, Complex)

# ↑↑↑↑↑↑↑↑↑↑↑↑ Necessary mathematic functions ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ To-string utilities ↓↓↓↓↓↓↓↓↓↓↓↓ #

num_format = "{:+.5}"


def set_n_digits(n):
    """
    Sets the number of digits "after the dot" you want to print.  
    For example if you call `set_n_digits(3)` and then `print(my_qubit)`  
    the result will be something like `'+0.707|0> +0.707|1>'`
    """
    if not isinstance(n, int): raise TypeError('The number of digits must be an int')
    if n<1: raise ValueError('Number of digits too low')
    global num_format
    num_format = "{:+."+str(n)+"}"


def round_float(strx):
    # round a float nubmer according to num_format
    for i in range(len(strx)-1, 0, -1):
        if strx[i] != '0':
            return strx[0:i] if strx[i]=='.' else strx[0:i+1]


def val2str(x):
    # generate a string of a complex value according to the number of digits
    if isinstance(x, complex):
        if equal(x.imag, 0):
            return round_float(num_format.format(x.real))
        elif equal(x.real, 0):
            return round_float(num_format.format(x.imag))+'j'
        else:
            return '+('+round_float(num_format.format(x.real)) + round_float(num_format.format(x.imag))+'j)'
    else:
        return round_float(num_format.format(x))


def formatProbs(state, symb):
    # used to formatting the probabilities of a qubit
    format_perc = lambda x: val2str(mod_square(x)*100).replace('+','')+'%'
    s=''
    for i, b in enumerate(symb):
        s += '|' + b + '>: ' + format_perc(state[i])+'\n'

    return s


# ↑↑↑↑↑↑↑↑↑↑↑↑ To-string utilities ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ From-string utilities ↓↓↓↓↓↓↓↓↓↓↓↓ #

regex = re.compile(r'([+\-]?)\s*([+\-0-9.j]*)\s*\|([^>]+)>')

def str2states(s):
    # transform a string into a dictionare {state : value}
    states = defaultdict(int)
    for sp in re.finditer(regex, s):
        #NB: sign=1, value=2, state=3
        if len(sp[2])==0:
            states[sp[3]] += -1 if sp[1]=='-' else 1
        else:
            states[sp[3]] += complex(sp[1]+sp[2])

    return states


def states2list(states, symbols):
    # - calssical usage: symbols = Basis.symbols
    # - multi-qubit usage: symbols = Vdigit([all bases])

    return [states.get(symb, 0) for symb in symbols]


# ↑↑↑↑↑↑↑↑↑↑↑↑ From-string utilities ↑↑↑↑↑↑↑↑↑↑↑↑ #



# ↓↓↓↓↓↓↓↓↓↓↓↓ Vdigit classes ↓↓↓↓↓↓↓↓↓↓↓↓ #


class Vdigit: # Variable Base digits
    """
    Transform a number into a representation digit-base:  
    every digit of this representations has his base and it's independent of other digits.

    Usage: 
    + `bases`: a list that describe the length of basis for each digit (if it's a list of int) and the symbols that digit can assume (if it's a list fo strings)
    + `start_value` (optional): the initial value

    Example:  
    + `Vdigit([1,2,3], 5) -> '012'`  
    + `Vdigit(['abc','01234'], 10) -> 'c0'`
    """

    def __init__(self, bases, start_value = 0):
        self.ds = []
        self.bases = bases
        self.hasNext = True

        if isinstance(bases[0], int):
            self.symb = lambda i: STD_SYMBOLS
            self.dim = lambda i: self.bases[i]
        else:
            self.symb = lambda i: self.bases[i]
            self.dim = lambda i: len(self.bases[i])

        self.convert(start_value)


    def convert(self, value):
        if value<0: raise ValueError('The value must be greater or equal to 0')

        self.ds = []
        i = len(self.bases)-1
        while value:
            self.ds.append( value%self.dim(i) )
            value = value//self.dim(i)
            i -= 1

        self.ds.reverse()

        if len(self.ds) < len(self.bases):
            self.ds = [0 for _ in range(len(self.bases)-len(self.ds))] + self.ds
        elif len(self.ds) > len(self.bases):
            raise ValueError('Value too high')


    def extract(self):
        "Extracts the integer value represented by this Vdigit"
        b_prod=1 ; val=0
        for i in range(len(self.ds)-1, -1, -1):
            val += self.ds[i]*b_prod
            b_prod *= self.dim(i)
        return val


    def __pp(self, i):
        if i < 0:
            self.hasNext = False
        
        self.ds[i]+=1

        if self.ds[i] >= self.dim(i):
            self.ds[i] = 0
            self.__pp(i-1)


    def _pp_ent(self, i, j, x):
        # custom method used to swap an entanged state
        if x < 0:
            self.hasNext = False
            return
        elif x in (i,j):
            x -= 1
        else:
            self.ds[x] += 1
            if self.ds[x] >= self.dim(x):
                self.ds[x]=0
                x -= 1
            else:
                return

        self._pp_ent(i,j,x)


    def pp(self):
        """
        The "plus plus" method: do a "+1" to the value of this Vdigit
        
        If it's is out of digits raises a StopIteration exception
        """
        if not self.hasNext:
            raise StopIteration

        self.__pp(len(self.bases)-1)
        

    def __iter__(self):
        return self


    def __next__(self):
        out = str(self)
        self.pp() #++
        return out


    def revStr(self):
        """
        Returns the reversed string of the representation  
        (for example if the representation is '1a', this function returns 'a1')
        """
        return "".join(reversed(str(self)))
    

    def __str__(self):
        return "".join( (self.symb(i)[self.ds[i]] for i in range(len(self.bases))) )

# ↑↑↑↑↑↑↑↑↑↑↑↑ Vdigit classes ↑↑↑↑↑↑↑↑↑↑↑↑ #
